# encoding: utf-8
import hashlib
import logging
from copy import copy
from slimurl import URL
from tornado.gen import coroutine, Return
from tornado.httpclient import AsyncHTTPClient
from tornado.ioloop import IOLoop
from tornado.locks import Lock
from tornado_xmlrpc.client import ServerProxy
from pypi_server.cache import Cache, HOUR, MONTH
from pypi_server.hash_version import HashVersion


log = logging.getLogger(__name__)


def normalize_package_name(name):
    return name.lower().replace("_", "-")


class PYPIClient(object):
    CLIENT = None
    BACKEND = None
    THREAD_POOL = None
    INDEX = None
    XMLRPC = None
    LOCK = None

    @classmethod
    def configure(cls, backend, thread_pool):
        cls.CLIENT = AsyncHTTPClient(io_loop=IOLoop.current())
        cls.BACKEND = backend
        cls.THREAD_POOL = thread_pool
        cls.XMLRPC = ServerProxy(
            str(copy(backend)(path="/pypi")),
        )
        cls.LOCK = Lock()

    @classmethod
    @coroutine
    @Cache(HOUR, files_cache=True, ignore_self=True)
    def packages(cls):
        with (yield cls.LOCK.acquire()):
            index = dict(
                map(
                    lambda x: (x.lower().replace("_", "-"), x),
                    (yield cls.XMLRPC.list_packages())
                )
            )

            log.info("Remote PYPI index updated: %d packages", len(index))
            raise Return(index)

    @classmethod
    @coroutine
    @Cache(4 * HOUR, files_cache=True, ignore_self=True)
    def search(cls, names, descriptions, operator="or"):
        assert operator in ('or', 'and')
        result = yield cls.XMLRPC.search({'name': names, 'description': descriptions}, operator)
        raise Return(result)

    @classmethod
    @coroutine
    def exists(cls, name):
        try:
            real_name = yield cls.find_real_name(name)
        except LookupError:
            raise Return(False)

        releases = yield cls.releases(real_name)
        if not releases:
            raise Return(False)

        raise Return(True)

    @classmethod
    @coroutine
    def find_real_name(cls, name):
        name = normalize_package_name(name)

        packages = yield cls.packages()
        real_name = packages.get(name.lower())

        if real_name is None:
            raise LookupError("Package not found")

        raise Return(real_name)

    @classmethod
    @coroutine
    @Cache(4 * HOUR, files_cache=True, ignore_self=True)
    def releases(cls, name):
        process_versions = lambda x: set(map(HashVersion, x))

        all_releases, current_releases = yield [
            cls.XMLRPC.package_releases(name, True),
            cls.XMLRPC.package_releases(name)
        ]

        all_releases = process_versions(all_releases)
        current_releases = process_versions(current_releases)

        hidden_releases = all_releases - current_releases

        res = []
        for x in current_releases:
            x.hidden = False
            res.append(x)

        for x in hidden_releases:
            x.hidden = True
            res.append(x)

        raise Return(set(res))

    @classmethod
    @coroutine
    @Cache(MONTH, files_cache=True, ignore_self=True)
    def release_data(cls, name, version):
        info, files = yield [
            cls.XMLRPC.release_data(str(name), str(version)),
            cls.XMLRPC.release_urls(str(name), str(version))
        ]

        download_url = info.get('download_url')
        if download_url and not files:
            try:
                url = URL(download_url)
                filename = url.path.split('/')[-1]

                if "#" in filename:
                    filename = filename.split("#")[0]

                response = yield cls.CLIENT.fetch(download_url)

                files = [{
                    'filename': filename,
                    'md5_digest': hashlib.md5(response.body).hexdigest(),
                    'downloads': -1,
                    'url': download_url,
                    'size': len(response.body),
                    'comment_text': None,
                }]
            except Exception as e:
                files = []
                log.error("Error when trying to download version %s of package %s", version, name)
                log.exception(e)

        else:
            files = sorted(
                files,
                key=lambda x: x['filename']
            )

        raise Return((info, files))



# encoding: utf-8
import logging
from tornado.gen import coroutine, maybe_future
from tornado.ioloop import IOLoop
from tornado.web import HTTPError
from pypi_server.handlers.base import BaseHandler, threaded


try:
    import ujson as json
except ImportError:
    import json


log = logging.getLogger(__name__)


def split_header(header):
    return dict(
        map(
            lambda x: list(map(lambda y: y.strip('"\''), x.split("=", 1))),
            filter(
                lambda x: "=" in x,
                map(lambda x: x.strip(), header.split(";"))
            )
        )
    )


class JSONHandler(BaseHandler):
    __slots__ = ("_json", "__io_loop")

    @coroutine
    def prepare(self, *args, **kwargs):
        self.__io_loop = IOLoop.current()

        if self.request.method.upper() in ('POST', 'PUT'):
            content_type = self.request.headers.get('Content-Type', '')

            if 'application/json' not in content_type:
                raise HTTPError(415)

            charset = split_header(content_type).get('charset', 'utf-8')

            try:
                self._json = yield self.thread_pool.submit(
                        self._from_json,
                        self.request.body.decode(charset)
                )
            except Exception as e:
                log.exception(e)
                raise HTTPError(415)
        else:
            self._json = None

    @classmethod
    @threaded
    def _to_json(cls, data):
        return json.dumps(data)

    def response(self, data):
        self.set_header('Content-Type', 'application/json')
        self._auto_finish = False

        self.__io_loop.add_callback(self._async_response, data)

    @coroutine
    def _async_response(self, data):
        data = yield maybe_future(data)
        resp = yield self._to_json(data)

        if not self._finished:
            log.debug("Sending: %r", resp)
            self.finish(resp)

    @staticmethod
    def _from_json(data):
        return json.loads(data)

    @property
    def json(self):
        return self._json


import pypi_server.handlers.api.login
import pypi_server.handlers.api.users
import pypi_server.handlers.api.user
import pypi_server.handlers.api.packages
import pypi_server.handlers.api.package
import pypi_server.handlers.api.version



# encoding: utf-8
import logging
from functools import wraps
from tornado.gen import Return
from tornado.web import RequestHandler, HTTPError

try:
    import cPickle as pickle
except ImportError:
    import pickle

try:
    import ujson as json
except ImportError:
    import json


log = logging.getLogger(__name__)


class BaseHandler(RequestHandler):
    _NULL = object()
    THREAD_POOL = None
    STORAGE = None

    @property
    def thread_pool(self):
        return self.THREAD_POOL

    @staticmethod
    def _log_result(function):
        def logger(result):
            exc = result.exception()
            if exc:
                if not isinstance(exc, (Return, HTTPError)):
                    log.exception(exc)
            else:
                log.debug(
                    "Result of function %s is: %r",
                    function.__name__,
                    result.result()
                )

        return logger

    @classmethod
    def threaded(cls, func):
        @wraps(func)
        def wrap(*args, **kwargs):
            f = cls.THREAD_POOL.submit(func, *args, **kwargs)
            f.add_done_callback(cls._log_result(func))
            return f

        return wrap

    def get_secure_cookie(self, name, value=None, max_age_days=31, min_version=None):
        val = super(BaseHandler, self).get_secure_cookie(
            name, None, max_age_days=max_age_days, min_version=min_version
        )

        if val is None:
            return value

        return pickle.loads(val) if val else val

    def set_secure_cookie(self, name, value, expires_days=30, version=None, **kwargs):
        return super(BaseHandler, self).set_secure_cookie(
            name, pickle.dumps(value, protocol=2), expires_days=expires_days,
            version=version, **kwargs
        )


threaded = BaseHandler.threaded


# encoding: utf-8
import logging
import os
import signal
from time import time
from functools import wraps
from inspect import isgeneratorfunction
from collections import defaultdict
from multiprocessing import RLock
from tornado.gen import Return, sleep
from tornado.ioloop import IOLoop
from tornado.locks import Lock


try:
    import cPickle as pickle
except ImportError:
    import pickle


FunctionType = type(lambda: None)
log = logging.getLogger("cache")


class Result(object):
    def __init__(self, result):
        self.__result = result
        self.__ts = time()

    @property
    def result(self):
        return self.__result

    @property
    def ts(self):
        return self.__ts


class Cache(object):
    __slots__ = ('timeout', 'ignore_self', 'oid', 'files_cache')

    CACHE_DIR = None
    CACHE = {}
    FUTURE_LOCKS = defaultdict(Lock)
    RLOCKS = defaultdict(RLock)

    def __init__(self, timeout, ignore_self=False, oid=None, files_cache=False):
        self.timeout = timeout
        self.ignore_self = ignore_self
        self.oid = oid
        self.files_cache = files_cache

    @staticmethod
    def hash_func(key):
        if isinstance(key, FunctionType):
            return ".".join((key.__module__, key.__name__))
        else:
            return str(key)

    @classmethod
    def invalidate(cls, func):
        fkey = cls.hash_func(func)

        hash_fkey = hash(fkey)
        for key in filter(lambda x: hash(x[0]) == hash_fkey, cls.CACHE):
            log.debug('INVALIDATING Cache for %r', key)
            cls.CACHE.pop(key, -1)
            cls.FUTURE_LOCKS.pop(key, -1)
            cls.RLOCKS.pop(key, -1)

    def get_cache(self, key):
        if self.files_cache:
            fname = self.get_cache_file(key)

            if not os.path.exists(fname):
                return None

            with open(fname, 'rb') as f:
                result = pickle.load(f)

            if result.ts < (time() - self.timeout):
                IOLoop.current().add_callback(os.remove, fname)
                return None

            return result

        return self.CACHE.get(key)

    def set_cache(self, key, value):
        if self.files_cache:
            fname = self.get_cache_file(key)

            with open(fname, 'wb+') as f:
                pickle.dump(value, f)

        else:
            self.CACHE[key] = value

    @classmethod
    def invalidate_all(cls, *args, **kwargs):
        log.warning("Invalidating all memory cache.")
        cls.CACHE.clear()
        cls.FUTURE_LOCKS.clear()
        cls.RLOCKS.clear()

        log.warning("Invalidating all disk cache.")
        files = filter(
            lambda x: os.path.isfile(x),
            (os.path.join(cls.CACHE_DIR, f) for f in os.listdir(cls.CACHE_DIR))
        )

        for file in files:
            try:
                os.remove(file)
            except Exception as e:
                log.exception(e)

    def __call__(self, func):
        is_generator = isgeneratorfunction(func)
        key = self.oid or self.hash_func(func)

        def get_hash(func, args, kwargs):
            return tuple(
                map(
                    hash,
                    (
                        key,
                        tuple(map(hash, args[1:] if self.ignore_self else args)),
                        tuple(map(lambda x: tuple(map(hash, x)), kwargs.items())),
                        is_generator,
                    )
                )
            )

        @wraps(func)
        def wrap(*args, **kwargs):
            io_loop = IOLoop.current()

            args_key = get_hash(func, args, kwargs)
            start_time = io_loop.time()

            with self.RLOCKS[args_key]:
                ret = self.get_cache(args_key)

                if isinstance(ret, Result):
                    log.debug("HIT Cache [%s] %r", key, args_key)
                    return ret.result

                ret = Result(func(*args, **kwargs))
                self.set_cache(args_key, ret)

                io_loop.add_callback(
                    io_loop.call_later,
                    self.timeout,
                    self._expire,
                    key,
                    args_key
                )

                log.debug(
                    "MISS Cache [%s] %r. Execution time %.6f sec.",
                    key,
                    args_key,
                    io_loop.time() - start_time
                )
                return ret.result

        @wraps(func)
        def wrap_gen(*args, **kwargs):
            io_loop = IOLoop.current()

            args_key = get_hash(func, args, kwargs)
            start_time = io_loop.time()

            with (yield self.FUTURE_LOCKS[args_key].acquire()):
                ret = self.get_cache(args_key)

                if isinstance(ret, Result):
                    yield sleep(0)
                    log.debug("HIT Cache [%s] %r", key, args_key)
                    raise Return(ret.result)

                gen = func(*args, **kwargs)

                try:
                    f = next(gen)
                    while True:
                        try:
                            res = yield f
                            f = gen.send(res)
                        except (Return, StopIteration):
                            raise
                        except Exception as e:
                            f = gen.throw(e)

                except Return as e:
                    ret = Result(e.value)
                except StopIteration as e:
                    ret = Result(getattr(e, 'value', None))

                if ret.result:
                    self.set_cache(args_key, ret)

                    io_loop.add_callback(
                        io_loop.call_later,
                        self.timeout,
                        self._expire,
                        key,
                        args_key
                    )

                    log.debug(
                        "MISS Cache [%s] %r. Execution time %.6f sec.",
                        key,
                        args_key,
                        io_loop.time() - start_time
                    )
                else:
                    log.warning(
                        "Generator '%s' no return any value. Cache ignoring.",
                        self.hash_func(func)
                    )

                    log.debug("INVALID Cache [%s] %r", key, args_key)

                raise Return(ret.result)

        return wrap_gen if is_generator else wrap

    def get_cache_file(self, args_key):
        return os.path.join(self.CACHE_DIR, str(hash(args_key)))

    def _expire(self, key, args_key):
        if self.files_cache:
            fname = self.get_cache_file(args_key)
            if os.path.exists(fname):
                os.remove(fname)

        self.CACHE.pop(args_key, -1)
        log.debug("EXPIRED Cache [%s] %r", key, args_key)


signal.signal(signal.SIGUSR1, Cache.invalidate_all)


MINUTE = 60
HOUR = MINUTE * 60
DAY = 24 * HOUR
WEEK = 7 * DAY
MONTH = 4 * WEEK
